; forth.asm -- a tiny FORTH compiler/interperter
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; Data stack:
;	EBP is the data stack pointer
;	It's a push-down stack, just like the CALL/RET stack
;	Use the DPUSH/DPOP macros
;	Also use S(N) to access the Nth element on the stack
;   *** Note: You can't use PUSHA/POPA because those affect EBP! ***
;
; Bugs:
;
; To do:
;
;	Implement wordlists: Wordlists will be normal Retro objects. 
;	FIND_WORD will isearch() under each object in the search order.
;
;	Allocate variables on a stack frame instead of in .data/.bss, so
;	it's reentrant.
;
;	Case-insensitive search for words (as an environment option)
;
;	Allow IF, CASE, and loops in interpretation mode (these would act
;	like C preprocessor directives: #ifdef, #endif, etc.)
;
;	Test everything out!
;

%define	INBUF_SIZE	80	;Size of input buffer (command line)


	SECTION .text
init_forth:
	sprint	"retroForth starting up...",10

	mov	esi, table
.loop	mov	ebx, [esi]
	or	ebx, ebx
	jz	.done
	dpush	esi		; name
	add dword S(0), byte 16
	mdpush	[esi+12]	; name length
	mdpush	[esi]		; address
	mdpush	[esi+4]		; size
	mdpush	[esi+8]		; flags
	call	insert		;insert it

	add	esi, 16		;skip past name, to next entry
.1	inc	esi
	cmp byte [esi-1], 0
	jne	.1
	jmp short .loop
.done	ret



BOOTSTRAP:
	;; Get ready
;	lea	eax, [esp-4]
;	mov	[initial_esp], eax	;used by BYE

	sprint	"Running Forth bootstrap program...",10
	DPUSH	bforth
	call	EVALUATE
%ifdef BETA
        ;sprint "Running vga.fo...",10
        ;DPUSH  vga
        ;call   EVALUATE
        sprint "Running mouse.fo...",10
        DPUSH  mouse
        call   EVALUATE
%endif
	ret


forth:
	sprint	10,"Welcome to retroForth.",10
	call	INTERPRET
	ret


BYE:
	sprint	"Try turning off your computer :)",10
;	mov	esp, [initial_esp]
	ret
eBYE


;------------------- Parsing & Compilation --------------------

INTERPRET:	;(The top-level interpreter loop)
	cmp cell [_STATE], 0
	jnz	.skip
	sprint	" ok",10
.skip	mov	[_TO_IN], cell 0
	push cell [_INBUF]
	DPUSH	eax	;dummy
	pop cell S(0)
	DPUSH	INBUF_SIZE
	call	ACCEPT
	DPOP	eax
	mov	[_INBUF_LEN], eax
	call	PARSE_INBUF
	jmp	INTERPRET
eINTERPRET


EVALUATE:	;; ( string -- , interpret from a string )
		;; Note: nonstandard behavior (no 'count' parameter)
	mpush	ecx, esi
	mpush	cell [_BLK], cell [_SOURCE_ID], cell [_INBUF], cell [_INBUF_LEN], cell [_TO_IN]
	mov	[_BLK], cell 0
;	mov	[_SOURCE_ID], cell -1

	;; Now feed it to PARSE_INBUF line-by-line
	DPOP	esi
.1	mov	[_INBUF], esi
	xor	ecx, ecx
	mov	[_TO_IN], ecx
.2	cmp	[esi], byte 10	;scan for CR
	je	.cr
	cmp	[esi], byte 0	;scan for null
	je	.done
	inc	esi
	inc	ecx
	jmp short .2

.cr	mov	[esi], byte 0
	inc	esi
	mov	[_INBUF_LEN], ecx
;	pstr	[_INBUF]		;Debuging output
;	cr				;...
	call	PARSE_INBUF
	jmp short .1

.done	mpop	cell [_BLK], cell [_SOURCE_ID], cell [_INBUF], cell [_INBUF_LEN], cell [_TO_IN]
	mpop	ecx, esi
	ret
eEVALUATE


PARSE_INBUF:
	push	eax
.loop	call	PARSE_WORD
	cmp	S(0), cell 0
	jne	.find		;end of input? (length=0)
	DROP2
	jmp	.end

.find	call	FIND_WORD	;Search
	DPOP	eax
	or	eax, eax
	jz	.num		;0 = search failed, might be a number
	jns	.imm		;1 = immediate word (don't compile)

	cmp	[_STATE], cell 0	;Are we compiling?
	jnz	.comp

.imm	DPOP	eax		;Execute the word
	call	eax
	jmp short .loop

.comp	call	COMPILE		;Compile a 'CALL xt' instruction
	jmp short .loop

.num	call	TO_NUMBER	;Returns w/ number on stack,
	jc	.err		;or carry set on failure
	cmp	[_STATE], cell 0	;Compiling?
	je near .loop			;no
	call	LITERAL			;yes
	jmp short .loop

.err	cr
	pchr	'"'
	call	TYPE
	pchr	'"'
	sprint " undefined",10
	;; Error recovery: (change if desired!)
	mov	[_STATE], cell 0	;Reset state
;	mov	[_SOURCE_ID], cell 0	;Get input from console
	mov	ebp, FORTH_STACK	;Clear stack
.end	pop	eax
	ret
ePARSE_INBUF


PARSE_WORD:	; ( "name" -- c-addr u , parse a space-delimited word )
	mpush	eax, ebx, esi
	mov	ebx, [_INBUF]
	add	ebx, [_TO_IN]		;ebx = position in string
	mov	eax, [_INBUF_LEN]
	sub	eax, [_TO_IN]		;eax = characters remaining

	mov     esi, -1			;Skip whitespace
.skip
	inc     esi
	cmp     esi, eax
	jge     .s1
	cmp     byte [ebx+esi], 0x20    ;SPACE
	je      .skip
	cmp     byte [ebx+esi], 0x09    ;TAB
	je      .skip
	cmp     byte [ebx+esi], 0x0D    ;CR
	je      .skip
	cmp     byte [ebx+esi], 0x0A    ;LF
	je      .skip
.s1	add     [_TO_IN], esi
	sub     eax, esi
	add	ebx, esi
	DPUSH	ebx			;Address

        sub     esi, esi		;Count length
.length
        cmp     esi, eax                ;End of input?
        jge     .end
        cmp     byte [ebx+esi], 0       ;NULL
        je      .end
        cmp     byte [ebx+esi], 0x20    ;SPACE
        je      .end
        cmp     byte [ebx+esi], 0x09    ;TAB
        je      .end
        cmp     byte [ebx+esi], 0x0D    ;CR
        je      .end
        cmp     byte [ebx+esi], 0x0A    ;LF
        je      .end
        inc     esi
        jmp     .length

.end	DPUSH	esi             	;Length
	inc	esi
	add	[_TO_IN], esi
	mpop	eax, ebx, esi
	ret
ePARSE_WORD


FIND_WORD:	; ( c-addr u -- c-addr u 0 | xt 1 | xt -1 )
;;
;; Return flags:
;;	 0 = not found (string is left on stack)
;;	 1 = immediate
;;	-1 = normal
;;
	mpush	eax, ebx

	call	_2DUP
	call	isearch
	DPOP	eax	;address (xt)
	or	eax, eax	;matched?
	jz	.error
	DPOP	ebx	;flags
	DROP		;size (ignored)

.found	DROP2
	DPUSH	eax
	test	ebx, F_IMM	;examine flags
	jz	.norm
.imm	DPUSH	1
	jmp short .done
.norm
	DPUSH	-1
.done
	mpop	eax, ebx
	ret
.error
	DPUSH	eax
	jmp short .done
eFIND_WORD


;-------------------------- Text Output ---------------------------


TYPE:		;( c-addr u -- , display a counted string )
	mpush	eax, ecx, esi
	DPOP	ecx	;Count
	DPOP	esi	;Address
.loop	lodsb
	or	al, al	;0?
	jz	.end
	pchr	eax
	loop	.loop
.end	mpop	eax, ecx, esi
	ret
eTYPE


PRINT:		;( n -- , diplay a number )
	call print_num
	space
	ret
ePRINT


CR:	;; ( -- , carriage return )
	cr
	ret
eCR


ACCEPT:		;( c-addr +n1 -- +n2 , read line into c-addr )
	mpush	S(1), S(0)
	DROP
	call	readline
	mpop	S(0)
	add	esp, 1*4
	ret
eACCEPT:


KEY:		;( -- char , read a character from keyboard )
	push	eax
	call	getc
	movzx	eax, al
	DPUSH	eax
	pop	eax
	ret
eKEY


TO_NUMBER:	;( c-addr u -- n|u )
;; TRY to convert to a number
	;
	; Algorithm:
	;	Check for - and set flag
	;	Read a digit
	;	Multiply accumulator by base & add digit
	;
	mpush	eax, ebx, ecx, edx, esi, edi
	DPOP	edi	; length (ignored)
	DPOP	esi	; string
	xor	eax, eax		;eax = Counter (the number)
	xor	ebx, ebx		;clear ebx (ebx is added to eax)
	xor	ecx, ecx		;ecx = # of digits
	mov	[sign], al		;sign = 0 (-1 means negative)

	cmp	[esi], byte '-'		;Negative?
	jne	.loop
	dec	byte [sign]
	inc	esi

.loop
	mov	bl, [esi]	;Fetch next digit & translate it
	cmp	bl, '0'
	jb	.number_end		;still here? bl >= '0'
	cmp	bl, '9'			;'0' <= bl <= '9'
	jbe	.decimal
	cmp	bl, 'A'
	jb	.number_end		;still here? bl >= 'A'
	cmp	bl, 'Z'			;'A' <= bl <= 'Z'
	jbe	.upr
	cmp	bl, 'a'
	jb	.number_end
	cmp	bl, 'z'
	jbe	.lwr
	jmp short .number_end

.lwr	sub	bl, 'a'-10		;Convert lowercase letter to #
	jmp short .check_base
.upr	sub	bl, 'A'-10		;Convert uppercase letter to #
	jmp short .check_base
.decimal	sub	bl, '0'			;Convert decimal digit to #
.check_base	cmp	bl, byte [_BASE]	;Make sure it's not above the base
	jae	.number_end

	mul	dword [_BASE]		;Add to the counter
	add	eax, ebx
	inc	ecx			;increment digit-counter
	inc	esi			;Get next digit
	jmp short .loop

.number_end
	or	ecx, ecx		;Any digits found?
	jz	.not_number
	cmp	byte [sign], 0		;Negative? flip sign.
	jz	.not_neg
	neg	eax
.not_neg
	DPUSH	eax		;return the number
	jmp short .end

.not_number
	DPUSH	esi
	DPUSH	edi
	stc
.end
	mpop	eax, ebx, ecx, edx, esi, edi
	ret
eTO_NUMBER


BASE:		;; variable
	DPUSH	_BASE
	ret
eBASE


;----------------------------- Miscellaneous ------------------------

PARSE:		;; ( char "lsjdfa<char>" -- c-addr u , parse a word delimited by <char> )
	mpush	eax, ebx, edx, esi
	DPOP	edx
	mov	ebx, [_INBUF]
	add	ebx, [_TO_IN]		;ebx = position in string
	mov	eax, [_INBUF_LEN]
	sub	eax, [_TO_IN]		;eax = characters remaining

	mov	esi, -1			;Skip whitespace
.skip
	inc     esi
	cmp     esi, eax
	jge     .s1
	cmp     byte [ebx+esi], dl
	je      .skip
.s1	add     [_TO_IN], esi
	sub     eax, esi
	add	ebx, esi
	DPUSH	ebx			;Address

        sub     esi, esi		;Count length
.length
        cmp     esi, eax                ;End of input?
        jge     .end
        cmp     byte [ebx+esi], 0       ;NULL
        je      .end
        cmp     byte [ebx+esi], dl    ;SPACE
        je      .end
        inc     esi
        jmp     .length

.end	DPUSH	esi             	;Length
	inc	esi
	add	[_TO_IN], esi
	mpop	eax, ebx, edx, esi
	ret
ePARSE


FETCH:		;; '@' ( addr -- x , read memory location )
	push	eax
	mov	eax, S(0)
	mov	eax, [eax]
	mov	S(0), eax
	pop	eax
	ret
eFETCH


STORE:		;; '!' ( x addr -- , write memory location )
	mpush	eax, ebx
	DPOP	ebx
	DPOP	eax
	mov	[ebx], eax
	mpop	eax, ebx
	ret
eSTORE


PLUS_STORE:	;; '+!' ( n addr -- , add n to number at addr )
	mpush	eax, ebx
	DPOP	ebx
	DPOP	eax
	add	[ebx], eax
	mpop	eax, ebx
	ret
ePLUS_STORE


C_FETCH:	;; 'C@' ( addr -- char , read memory location )
	push	eax
	mov	eax, S(0)
	movzx	eax, byte [eax]
	mov	S(0), eax
	pop	eax
	ret
eC_FETCH


C_STORE:	;; 'C!' ( char addr -- , write memory location )
	mpush	eax, ebx
	DPOP	ebx
	DPOP	eax
	mov	[ebx], al
	mpop	eax, ebx
	ret
eC_STORE


W_FETCH:	;; 'W@' ( addr -- word , read 16-bit memory location )
	push	eax
	mov	eax, S(0)
	movzx	eax, word [eax]
	mov	S(0), eax
	pop	eax
	ret
eW_FETCH


W_STORE:	;; 'W!' ( word addr -- , write 16-bit memory location )
	mpush	eax, ebx
	DPOP	ebx
	DPOP	eax
	mov	[ebx], ax
	mpop	eax, ebx
	ret
eW_STORE


IOR:	;; ( ioaddr -- iovalue , "in iovalue, ioaddr" )
	mpush	eax, edx
	mov	edx, S(0)
	xor	eax, eax
	in	al, dx
	mov	S(0), eax
	mpop	eax, ecx
	ret
eIOR


IOW:	;; ( iovalue ioaddr -- , "out ioaddr, iovalue" )
	mpush	eax, edx
	DPOP	edx
	DPOP	eax
	out	dx, al
	mpop	eax, edx
	ret
eIOW


_INT:	;; ( n -- , "int n" in assembly )
	push	eax
	DPOP	eax
	mov	[.n], al
	db	0xCD	; INT
.n	db	0	; n
	pop	eax
	ret
e_INT


ALIGNED:	;; ( addr -- a-addr )
	test	byte S(0), 3	;aligned if first 2 bits are 0
	jz	.end
	and	byte S(0), ~3
	add	dword S(0), byte 4
.end	ret
eALIGNED


_ALIGN:	;; ( -- , aligns "HERE" )
	test	byte [_HERE], 3	;aligned if first 2 bits are 0
	jz	.end
	and	byte [_HERE], ~3
	add	dword [_HERE], byte 4
.end	ret
e_ALIGN


HERE:		;; ( -- addr , get heap address )
	push	eax
	mov	eax, [_HERE]
	DPUSH	eax
	pop	eax
	ret
eHERE


COMMA:		; ',' ( x -- , moves a cell from stack to heap )
	mpush	eax, edi
	DPOP	eax
	mov	edi, [_HERE]
	stosc
	mov	[_HERE], edi
	mpop	eax, edi
	ret
eCOMMA


C_COMMA:	;; 'C,' ( char --, moves a character from stack to heap )
	mpush	eax, edi
	DPOP	eax
	mov	edi, [_HERE]
	stosb
	mov	[_HERE], edi
	mpop	eax, edi
	ret
eC_COMMA


ALLOT:		;; ( n -- , allocate n bytes on heap )
	push	eax
	DPOP	eax
;pchr '+'
;pnum eax
	add	[_HERE], eax
	pop	eax
	ret
eALLOT


_CMOVE:		;; 'CMOVE' ( src dest count -- )
	mpush	ecx, esi, edi
	DPOP	ecx
	DPOP	edi
	DPOP	esi
	cld
	rep	movsb
	mpop	ecx, esi, edi
	ret
e_CMOVE


_CMOVE2:	;; 'CMOVE>' ( src dest count -- )
	mpush	ecx, esi, edi
	DPOP	ecx
	DPOP	edi
	DPOP	esi
	add	edi, ecx
	add	esi, ecx
	std
	rep	movsb
	cld
	mpop	ecx, esi, edi
	ret
e_CMOVE2


FILL:		;; ( addr count char -- , starting at addr, fill count bytes w/ char )
	mpush	eax, ecx, edi
	DPOP	eax
	DPOP	ecx
	DPOP	edi
	cld
	rep	stosb
	mpop	eax, ecx, edi
	ret
eFILL


CELL_PLUS:	;; 'CELL+' ( addr -- addr , adds CELLSIZE to addr )
	add cell S(0), CELLSIZE
	ret
eCELL_PLUS


TICK:		;; ' ( "name" -- xt , xt = address of word "name" )
	push	eax
	call	PARSE_WORD
	call	FIND_WORD
	DPOP	eax
	or	eax,eax		;Check for error
	jnz	.ok
	call	_2DROP
	DPUSH	eax		;Return 0 if not found
.ok	pop	eax
	ret
eTICK


EXECUTE:	;; ( xt -- , CALL xt )
	call	S(0)
	ret
eEXECUTE


LITERAL:	;; ( x -- , Compile machine-code for DPUSH x )
	DPUSH cell 0x83		; SUB ebp, byte CELLSIZE
	call	C_COMMA
	DPUSH cell 0xED
	call	C_COMMA
	DPUSH cell CELLSIZE
	call	C_COMMA
	DPUSH cell 0xC7		; MOV cell [ebp], x
	call	C_COMMA
	DPUSH cell 0x45
	call	C_COMMA
	DPUSH cell 0x00
	call	C_COMMA
	call	COMMA
	ret
eLITERAL


COMPILE:	;; '[COMPILE]' ( xt -- , generate a "CALL xt" instruction )
	DPUSH	cell 0xE8
	call	C_COMMA
	call	HERE
	call	CELL_PLUS
	call	MINUS
	call	COMMA
	ret
eCOMPILE


CONSTANT:	;; ( x "name" -- )
	push cell S(0)		;* would be unnecessary if the control stack was separate from data
	DROP			;*
	call	NEW_WORD
	DPUSH	eax   ;dummy	;*
	pop cell S(0)		;*
	call	LITERAL
	call	ADD_WORD
	ret
eCONSTANT


VARIABLE:	;; ( "name" -- )
	push cell [_HERE]	;*
	DPUSH cell CELLSIZE	; allocate space for variable
	call	ALLOT
	call	NEW_WORD
	DPUSH	eax   ;dummy	;*
	pop cell S(0)		;*
	call	LITERAL
	call	ADD_WORD
	ret
eVARIABLE


LEFT_BRACKET:	;; '[' ( enter interpretation state )
	mov cell [_STATE], -1
	ret
eLEFT_BRACKET


RIGHT_BRACKET:	;; ']' ( enter compilation state )
	mov cell [_STATE], 0
	ret
eRIGHT_BRACKET


START_COMPILE:	;; ':' ( "name" -- colon-sys )
	call	NEW_WORD
	mov cell [_STATE], -1
	ret
eSTART_COMPILE


END_COMPILE:	;; ';' ( colon-sys -- )
	mov cell [_STATE], 0
	call	ADD_WORD
	ret
eEND_COMPILE


NEW_WORD:	;; ( "name" -- colon-sys )
;; Prepare to define a new word (could be ':', 'variable', 'constant'...)
;; After the word is defined, ADD_WORD removes the 'colon-sys' stuff from
;; the stack and add the word to the dictionary.
;;
;; colon-sys = name addr
;;		name = c-addr u
	call	PARSE_WORD
	call	HERE
	ret
eNEW_WORD


ADD_WORD	;; ( colon-sys -- )
	mpush	eax, ebx, ecx, edx
	DPUSH cell 0xC3	; add RET instruction
	call	C_COMMA

	call	HERE
	call	OVER
	call	MINUS
	DPOP	ecx	; len= HERE - addr
	DPOP	eax	; addr
	mov	edx, S(0)	;namelen
	;; Build asciiz string (on heap)
	;; ( src count )
	call	HERE
	DPOP	ebx	; address of new string
	call	DUP
	call	_ROT	; ( count src count )
	call	HERE
	call	SWAP	; ( count src dest count )
	call	_CMOVE	; ( count )
	call	ALLOT
	DPUSH	0
	call	C_COMMA	; null terminator
	;; Add to object list
	dpush	ebx, edx, eax, ecx, T_FORTH
	call	insert
;TODO: set [_LAST]
	mpop	eax, ebx, ecx, edx
	ret
eADD_WORD


COMMENT:	;; '\' ( ignore everything till end of line )
	mpush	ebx
	mov	ebx, [_INBUF]
	add	ebx, [_TO_IN]		;ebx = position in string
	mov	[ebx], byte 0
	mpop	ebx
	ret
eCOMMENT


;------------------------------ Stack Manipulation -----------------------

DUP:	;; ( x -- x x , duplicate TOS )
	mpush	S(0)
	sub	ebp, byte CELLSIZE
	mpop	S(0)
	ret
eDUP


?DUP:	;; ( x -- x x , duplicate x if x<>0 )
	cmp CELL S(0), 0
	jz	.end
	call	DUP
.end	ret
e?DUP


_DROP:	;; ( x -- )
	DROP
	ret
e_DROP


_2DROP:	;; ( x -- )
	DROP2
	ret
e_2DROP


OVER:	;; ( x1 x2 -- x1 x2 x1 , duplicates S(1) )
	mpush	S(1)
	sub	ebp, byte CELLSIZE
	mpop	S(0)
	ret
eOVER


PICK:	;; ( ... N -- ... x , copy the Nth stack item [0 PICK = DUP] )
	mpush	eax
	mov	eax, S(0)
;	mov	eax, S(eax+1)
	mov	eax, [ebp + CELLSIZE*eax + CELLSIZE]
	mov	S(0), eax
	mpop	eax
	ret
ePICK


SWAP:	;; ( x1 x2 -- x2 x1 , swap the top 2 stack items )
	mpush	eax, ebx
	mov	eax, S(0)
	mov	ebx, S(1)
	mov	S(0), ebx
	mov	S(1), eax
	mpop	eax, ebx
	ret
eSWAP


_2SWAP:	;; ( a1 a2 b1 b2 -- b1 b2 a1 a2 , swap top 2 cell pairs )
	mpush	eax, ebx
	mov	eax, S(0)
	mov	ebx, S(2)
	mov	S(0), ebx
	mov	S(2), eax
	mov	eax, S(1)
	mov	ebx, S(3)
	mov	S(1), ebx
	mov	S(3), eax
	mpop	eax, ebx
	ret
e_2SWAP


_2OVER:	;; ( a1 a2 b1 b2 -- a1 a2 b1 b2 a1 a2 )
	mpush	S(2), S(3)
	sub	ebp, byte 2*CELLSIZE
	mpop	S(0), S(1)
	ret
e_2OVER


_2DUP:	;; '2DUP' ( x1 x2 -- x1 x2 x1 x2 , duplicates top 2 items )
	call	OVER
	call	OVER
	ret
e_2DUP


ROT:	;; 'ROT' ( x1 x2 x3 -- x2 x3 x1 , rotate left )
	mpush	eax, ebx
	mov	eax, S(0)
	mov	ebx, S(1)
	mov	S(1), eax
	mov	eax, S(2)
	mov	S(2), ebx
	mov	S(0), eax
	mpop	eax, ebx
	ret
eROT


_ROT:	;; '-ROT' ( x1 x2 x3 -- x3 x1 x2 , rotate right )
	mpush	eax, ebx
	mov	eax, S(2)
	mov	ebx, S(1)
	mov	S(1), eax
	mov	eax, S(0)
	mov	S(0), ebx
	mov	S(2), eax
	mpop	eax, ebx
	ret
e_ROT


ROLL:	;; ( xu xu-1 ... x0 u -- xu-1 ... x0 xu )
	mpush	eax, ebx, ecx
	DPOP	ecx
	mov	ebx, [ebp+ecx*CELLSIZE]
.l1	mov	eax, [ebp+ecx*CELLSIZE-CELLSIZE]
	mov	[ebp+ecx*CELLSIZE], eax
	loop	.l1
	mov	S(0), ebx
	mpop	eax, ebx, ecx
	ret
eROLL


R_FROM:
	copy_template .t, .te
	ret
.t	;; template code
	sub	ebp, byte CELLSIZE
	mpop	S(0)
.te
eR_FROM


R_FETCH:
	copy_template .t, .te
	ret
.t	;; template code
	sub	ebp, byte CELLSIZE
	mpop	S(0)
	sub	esp, byte CELLSIZE
.te
eR_FETCH


TO_R:
	copy_template .t, .te
	ret
.t	;; template code
	mpush	S(0)
	add	ebp, byte CELLSIZE
.te
eTO_R


DEPTH:	;; ( -- u , # of items on stack )
	mpush	eax
	mov	eax, FORTH_STACK
	sub	eax, ebp
	shr	eax, CELLSHIFT
	DPUSH	eax
	mpop	eax
	ret
eDEPTH


TRACE:	;; '.S' ( -- , show a stack trace )
	mpush	ebx, ecx
	pchr	'<'
	space
	call	DEPTH
	mov	ecx, S(0)
	call	PRINT
	pchr	'>'
	space
	or	ecx, ecx
	jz	.end
	mov	ebx, FORTH_STACK
.loop	sub	ebx, CELLSIZE
	MDPUSH	[ebx]
	call	print_num
	space
	loop	.loop
.end	cr
	mpop	ebx, ecx
	ret
eTRACE


;----------------------------- Math -------------------------------------

ABS:
	cmp	cell S(0), 0
	jge	.done
	neg	cell S(0)
.done	ret
eABS


MIN:
	push	eax
	DPOP	eax
	cmp	S(0), eax
	jle	.done
	mov	S(0), eax
.done	ret
eMIN


MAX:
	push	eax
	DPOP	eax
	cmp	S(0), eax
	jge	.done
	mov	S(0), eax
.done	ret
eMAX


NEGATE:
	neg	cell S(0)
	ret
eNEGATE


PLUS:
	push	eax
	DPOP	eax
	add	S(0), eax
	pop	eax
	ret
ePLUS


MINUS:
	push	eax
	DPOP	eax
	sub	S(0), eax
	pop	eax
	ret
eMINUS


_1PLUS:
	inc	cell S(0)
	ret
e_1PLUS


_1MINUS:
	dec	cell S(0)
	ret
e_1MINUS


_2MUL:
	shl	cell S(0), 1
	ret
e_2MUL


_2DIV:
	shr	cell S(0), 1
	ret
e_2DIV


MULTIPLY:
	mpush	eax, edx
	DPOP	eax
	mul	cell S(0)
	mov	S(0), eax
	mpop	eax, edx
	ret
eMULTIPLY


DIVIDE:
	mpush	eax, edx
	call	SWAP
	DPOP	eax
	ccd
	div	cell S(0)
	mov	S(0), eax
	mpop	eax, edx
	ret
eDIVIDE


MOD:
	mpush	eax, edx
	call	SWAP
	DPOP	eax
	ccd
	div	cell S(0)
	mov	S(0), edx
	mpop	eax, edx
	ret
eMOD


DIV_MOD:
	mpush	eax, edx
	mov	eax, S(1)
	ccd
	div	cell S(0)
	mov	S(1), edx
	mov	S(0), eax
	mpop	eax, edx
	ret
eDIV_MOD


UM_MOD:		;; 'um/mod' ( ud u1 - u2 u3 , unsigned double /MOD )
	mpush	eax, edx
	mov	eax, S(2)
	mov	edx, S(1)
	div	cell S(0)
	DROP
	mov	S(1), edx
	mov	S(0), eax
	mpop	eax, edx
	ret
eUM_MOD


MUL_DIV:
	mpush	eax, ebx, edx
	DPOP	ebx
	DPOP	eax
	mul	cell S(0)
	div	ebx
	mov	S(0), eax
	mpop	eax, ebx, edx
	ret
eMUL_DIV


MUL_DIV_MOD:
	mpush	eax, ebx, edx
	DPOP	ebx
	DPOP	eax
	mul	cell S(0)
	div	ebx
	mov	S(0), edx
	DPUSH	eax
	mpop	eax, ebx, edx
	ret
eMUL_DIV_MOD


M_MULTIPLY:	; 'M*' ( n1 n2 -- d , double-cell result )
	mpush	eax, edx
	mov	eax, S(0)
	mul	cell S(1)
	mov	S(1), eax
	mov	S(0), edx
	mpop	eax, edx
	ret
eM_MULTIPLY


;----------------------------- Logic -------------------------------------

BINAND:
     push eax
     DPOP eax
     and S(0),eax
     pop eax
     ret
eBINAND


BINOR:
     push eax
     DPOP eax
     or S(0),eax
     pop eax
     ret
eBINOR


BINXOR:
     push eax
     DPOP eax
     xor S(0),eax
     pop eax
     ret
eBINXOR


INVERT:		;( x1 -- x2 , 1's complement )
	not dword S(0)
	ret
eINVERT


LSHIFT:		;( x y -- x2 , shift x left by y bits )
	push	ecx
	DPOP	ecx
	shl dword S(0), cl
	pop	ecx
	ret
eLSHIFT


RSHIFT:		;( x y -- x2 , shift x right by y bits )
	push	ecx
	DPOP	ecx
	shr dword S(0), cl
	pop	ecx
	ret
eRSHIFT


;----------------------------- Tests -------------------------------------

EQUAL?:
     mpush eax,ebx
     DPOP eax
     DPOP ebx
     cmp eax,ebx
     jne .l1
     mov eax,-1
     jmp .l3
       .l1:
     xor eax,eax
       .l3:
     DPUSH eax
     mpop eax,ebx
     ret
eEQUAL?


LESS?:
     push eax
     DPOP eax
     cmp S(0),eax
     jnl .l1
     mov CELL S(0),-1
     jmp .l2
       .l1:
     mov CELL S(0),0
       .l2:
     pop eax
     ret
eLESS?


ULESS?:
     push eax
     DPOP eax
     cmp S(0),eax
     jnb .l1
     mov CELL S(0),-1
     jmp .l2
       .l1:
     mov CELL S(0),0
       .l2:
     pop eax
     ret
eULESS?


GREATER?:
     push eax
     DPOP eax
     cmp S(0),eax
     jng .l1
     mov CELL S(0),-1
     jmp .l2
       .l1:
     mov CELL S(0),0
       .l2:
     pop eax
     ret
eGREATER?


;-------------------------------------------------------------------------


IMMEDIATE:	;; Make the last word compiled "immediate"
;	callproc chflags, [_LAST], F_IMM
	ret
eIMMEDIATE

COMPILE_ONLY:	;; Make the last word compiled "compile-only"
;	callproc chflags, [_LAST], F_COM
	ret
eCOMPILE_ONLY


POSTPONE:	;; Compile the next word even if it's "immediate"
	call	PARSE_WORD	; c-addr u
	cmp	S(0), cell 0
	jnz	.find
	DROP2
	jmp short .end
.find	call	FIND_WORD	; c-addr u 0 | xt 1 | xt -1
	cmp	S(0), cell 0
;To do:
	; zero: Treat it as a number or error (like PARSE_INBUF.num)
	; nonzero: Compile 'xt'
.end	ret
ePOSTPONE


CREATE:		;; ( "name" -- )
	call	NEW_WORD
	call	HERE
	DPUSH	11	; size of a LITERAL and a RET
	call	PLUS
	call	LITERAL
	call	ADD_WORD
	ret
eCREATE


;To do:
DOES:
;
; 1. Compile a 'copy template' stub and a RET
; 2. Anything following DOES> and before ; goes into the template code.
; 3. At runtime, the template code is compiled into the current definition
;    (i.e., copied to HERE)
;
	ret
eDOES


;------------------------------ Branching ----------------------------------

AHEAD:		;; ( C: -- orig , orig = <forward reference> )
	call	HERE
	sub	S(0), byte CELLSIZE
	ret
eAHEAD


IF:		;; ( C: -- orig , orig = location of imm field in jz )
;;	Execution: ( x -- , skip following statements if x=0 )
;;
	copy_template .t, .te
	;; Calculate ORIG
	call	AHEAD			; orig
;	call	HERE
;	DPUSH	CELLSIZE
;	call	MINUS
	ret
.t	;; template code
	add	ebp, CELLSIZE
	cmp	[ebp-CELLSIZE], cell 0
	jz near 0			; <-- jump location
.te
eIF


ELSE:		;; ( C: orig1 -- orig2 )
	copy_template .t, .te		; compile: JMP <forward ref>
	call	HERE			; resolve orig1
	call	OVER
	add	S(0), byte CELLSIZE
	call	MINUS
	call	SWAP
	call	STORE
	call	AHEAD			; new forward ref: orig2
;	call	HERE
;	sub	S(0), byte CELLSIZE
	ret
.t	;; template code
	jmp	0
.te
eELSE


THEN:		;; ( C: orig -- , Resolve forward reference )
	call	HERE
	call	OVER
	call	CELL_PLUS
	call	MINUS
	call	SWAP
	call	STORE
	ret
eTHEN


;------------------------------- Looping ----------------------------------

BEGIN		;; ( C: -- dest )
	call	HERE			; dest
	ret
eBEGIN


WHILE		;; ( C: dest -- orig dest )
;;	Execution: ( x -- , break loop if x=0 )
;;		If x=0, jump to instructon after 'REPEAT'
;;	dest = address of BEGIN
;;	orig = address of forward reference in WHILE
;;
;;In forth,	: WHILE  POSTPONE IF  SWAP ;  IMMEDIATE
;;
	call	IF
	call	SWAP
	ret
eWHILE


REPEAT		;; ( C: orig dest -- )
;;	Execution: ( -- )
;;		Jump back to BEGIN
;;
;;In forth,	: REPEAT  POSTPONE AGAIN  POSTPONE THEN ;  IMMEDIATE
;;
	call	AGAIN
	call	THEN		;Resolve forward reference
	ret
eREPEAT


UNTIL		;; ( C: dest -- )
;;	Execution: ( x -- , Loop to dest if x=0 )
;;
	copy_template .t, .te
	;; change 'jz 0' to 'jz dest'
	call	HERE
	call	MINUS
	call	HERE			; relative address for jz
	sub	S(0), byte CELLSIZE	; address of imm field in jz
	call	STORE
	ret
.t	;; template code
	DROP
	cmp dword S(-1), byte 0
	jz near 0
.te
eUNTIL


AGAIN:		;; ( C: dest -- )
;;	Execution: ( x -- , Loop to dest unconditionally )
;;
	copy_template .t, .te
	;; change 'jmp 0' to 'jmp dest'
	call	HERE
	call	MINUS
	call	HERE			; relative address for jmp
	sub	S(0), byte CELLSIZE	; address of imm field in jmp
	call	STORE
	ret
.t	;; template code
	jmp	0
.te
eAGAIN


DO:		;; ( C: -- do-sys )
;;	Execution: ( limit first -- ) ( R: -- loop-sys )
;;
;; examples:
;;	<limit> <first> DO ... LOOP
;;	<limit> <first> DO ... <step> +LOOP
;;
	copy_template .t, .te
	call	HERE			; do-sys
	ret
.t	;; template
	mpush	S(1), S(0)	; limit, index (i)
	DROP2
.te
eDO


_LOOP:		;; ( C: do-sys -- )
;;	Execution: ( -- ) ( R: loop-sys1 --   | loop-sys2 )
;;
	copy_template .t, .te
	call	HERE
	sub	S(0), byte .te-.t1
	call	MINUS
	call	HERE				; relative address for jb
	sub	S(0), byte CELLSIZE+(.te-.t1)	; address of imm field in jb
	call	STORE
	ret
.t	;; template
	inc dword [esp]
	mov	eax, [esp]
	cmp	eax, [esp+CELLSIZE]
	jb near	0
.t1	add	esp, byte 2*CELLSIZE
.te
e_LOOP


PLOOP:		;; ( C: do-sys -- )
;;	Execution: ( -- ) ( R: loop-sys1 --   | loop-sys2 )
;;
	copy_template .t, .te
	call	HERE
	sub	S(0), byte .te-.t1
	call	MINUS
	call	HERE				; relative address for jb
	sub	S(0), byte CELLSIZE+(.te-.t1)	; address of imm field in jb
	call	STORE
	ret
.t	;; template
	DPOP	eax			; step
	add	[esp], eax
	mov	eax, [esp]
	cmp	eax, [esp+CELLSIZE]
	jb near	0
.t1	add	esp, byte 2*CELLSIZE
.te
ePLOOP


I:	;; ( -- n , copy 1st item [loop index] from return stack )
	push	eax
	mov	eax, [esp+2*CELLSIZE]
	DPUSH	eax
	pop	eax
	ret
eI


J:	;; ( -- n , copy 2nd item [2nd outermost loop index] from return stack )
	push	eax
	mov	eax, [esp+3*CELLSIZE]
	DPUSH	eax
	pop	eax
	ret
eJ


;-------------------------------------------------------------------------

;; Later...
; order/wordlist, disk blocks, file access (a/k/a raw objects), ...



	SECTION .data
;; System variables
_INBUF		dc	INBUF		;(might not need this)
_HERE		dc	FORTH_HEAP	;the address 'HERE' returns
base
_BASE		dc	10
_SOURCE_ID	dc	0
_BLK		dc	0
_STATE		dc	0
_LAST		dc	0		;last word added to dictionary

%include "words.inc"


bforth:	incbin	"bforth.fo"
        db 0
%ifdef BETA
mouse   incbin "mouse.fo"
        db 0
vga     incbin "vga.fo"
        db 0
%endif

	SECTION .bss
;; System variables
_INBUF_LEN	resc	1
_TO_IN		resc	1	; '>IN'

INBUF		resb	INBUF_SIZE

forth_hdr	resd	1	; location of 'root:forth' object header
sign		resb	1	; sign of # (for TO_NUMBER)
initial_esp	resd	1
